####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
import typing as tp
#import pandas as pd
import weakref as wref
from copy import deepcopy as dcopy

import _paramonte as pm
from paramonte.vis._BasePlot import BasePlot
from paramonte.vis.LineScatterPlot import LineScatterPlot
from paramonte.vis.DensityPlot import DensityPlot

Struct = pm.Struct
newline = pm.newline

####################################################################################################################################
#### GridPlot class
####################################################################################################################################

class GridPlot(BasePlot):
    """

    This is the GridPlot class for generating instances
    of histogram and contour figures in two and three dimensions
    based on a wide range of plotting tools from the matplotlib,
    seaborn, and other Python libraries.

    Normally, the public users are not supposed to use this class directly,
    although they can for the purposes other than plotting the ParaMonte
    simulation files.

        **Parameters**

            plotType

                A string indicating the name of the plot to be constructed.

            dataFrame (optional)

                A pandas dataFrame whose data will be plotted.

            methodName (optional)

                The name of the ParaMonte sample requesting the BasePlot.

            reportEnabled (optional)

                A boolean whose value indicates whether guidelines should be 
                printed in the standard output.

            resetPlot (optional)

                A function that resets the properties of the plot as desired 
                from outside. If provided, a pointer to this function will be
                saved for future internal usage.

        **Attributes**

            All of the parameters described above, are implicitly 
            stored in the object.

            columns

                An attribute that determines the columns of dataFrame to be 
                visualized. It can have three forms:

                    1.  A list of column indices in dataFrame.
                    2.  A list of column names in dataFrame.columns.
                    3.  A ``range(start,stop,step)`` of column indices.

                Examples:

                    1.  ``xcolumns = [0,1,4,3]``
                    2.  ``xcolumns = ["SampleLogFunc","SampleVariable1"]``
                    3.  ``xcolumns = range(17,7,-2)``

                The default behavior includes up to 5 columns of the dataFrame.

            ccolumn

                A string or integer that determines the column 
                of the dataFrame to serve as the color-values 
                corresponding to each point in the plot. 

                Examples:

                    1.  ``ccolumn = 7``
                    2.  ``ccolumn = "SampleLogFunc"``
                    3.  ``ccolumn = None``

                **NOTE**

                The value of ``ccolumns`` has precedence over all 
                other color properties of individual plotting tools, 
                unless ``ccolumns = None``.
                For example, any value of the ``c`` component of the ``scatter.kws``
                component of the ``LineScatterPlot`` object will be overridden by
                the specified value for ``ccolumns``.

            rows

                An attribute that determines the rows of dataFrame to be 
                visualized. It can be either:

                    1.  A ``range(start,stop,step)``, or, 
                    2.  A list of row indices in dataFrame.index.

                Examples:

                    1.  ``rows = range(17,7,-2)``
                    2.  ``rows = [i for i in range(7,17)]``

                The default behavior includes all rows of the dataFrame.

            set

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``set()``
                        function of the seaborn library should be made or not.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``set()`` function.

                        Example usage:

                            .. code-block:: python

                                set.kws.style = "darkgrid"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            axes (available only in 1D and 2D plots)

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``gca()`` function of the 
                        matplotlib library.

                        Example usage:

                            .. code-block:: python

                                axes.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            figure

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``figure()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``figure()`` function of 
                        the matplotlib library.

                        Example usage:

                            .. code-block:: python

                                figure.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            pairgrid

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``PairGrid()`` class of 
                        the seaborn library to generate a pairgrid.

                        Example usage:

                            .. code-block:: python

                                pairgrid.kws.dropna = True

                        **NOTE**

                        Passing some keywords, like ``corner`` attribute,
                        to the PairGrid will cause potential conflicts with
                        other attributes of ParaMonte's GridPlot. Use this
                        attribute with caution. That said, if a desired 
                        property is missing among the ``kws`` attributes, 
                        simply add the field and its value to the component.

            colorbar (exists only for plots that require colorbar)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``colorbar()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``colorbar()`` function of 
                        the matplotlib library.

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

                A colorbar will be added to a plot only if a color-mappings 
                is requested in the plot.

            plotType

                A structure with three attributes:

                    upper

                        A structure whose components determine what plots
                        can and should be added to the upper corner of 
                        the grid plot.

                    lower

                        A structure whose components determine what plots
                        can and should be added to the lower corner of 
                        the grid plot.

                    diag

                        A structure whose components determine what plots
                        can and should be added to the diagonal of 
                        the grid plot.

                Each of the above structures has three components:

                    enabled

                        A boolean that determines if the subplots in the
                        corresponding section of the grid plot should be 
                        added or not.

                    names

                        A list of strings that represent the types of plots
                        that can be added to the corresponding section of 
                        the grid plot. 

                    value

                        A string whose value can be one of the elements of 
                        the ``names`` attribute explained in the above. It 
                        determines the type of the subplot to be added to 
                        the corresponding section of the grid plot. 

                Example usage:

                    .. code-block:: python

                        plotType.upper.enabled = False # no upper corner plots
                        plotType.lower.value = "scatter" # add scatter to lower

            layout

                A structure whose components are directly passed to the 
                corresponding plotting tools of the ParaMonte library to 
                draw the corresponding subplots, if they are activated.

                Example usage:

                    .. code-block:: python

                        layout.contour.contour.kws.colors = "blue"

            currentFig

                A structure whose attributes are the outputs of various plotting 
                tools used to make the current figure. These include the handle 
                to the current figure, the handle to the current axes in the plot, 
                the handle to the colorbar (if any exists), and other Python 
                plotting tools used to make to generate the figure.

        **Returns**

            self

                An object of class ``GridPlot``.

    ---------------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , plotType      # : str
                , dataFrame     = None # : tp.Optional[ pd.DataFrame ] = None
                , methodName    = "ParaMonte" # : tp.Optional[ str ] = "ParaMonte"
                , reportEnabled = True # : tp.Optional[ bool ] = True
                , resetPlot     = None
                ):

        super().__init__( plotType = plotType
                        , dataFrame = dataFrame
                        , methodName = methodName
                        , reportEnabled = reportEnabled
                        , resetPlot = resetPlot
                        )

        self._reset()
        if resetPlot is None: self._resetPlot = self._reset
        self._progress.note()

    ################################################################################################################################
    #### _reset
    ################################################################################################################################

    def _reset(self):

        super()._reset()

        self.columns = None
        self.ccolumn = [] # use the default colormap data (which is the data count)

        self.pairgrid = Struct()
        self.pairgrid.kws = Struct()

        self.colorbar = Struct()
        self.colorbar.enabled = True
        self.colorbar.kws = Struct()

        ############################################################################################################################
        #### setup corner types
        ############################################################################################################################

        cornerNames =   [ "line"
                        , "scatter"
                        , "lineScatter"
                        , "contourf"
                        , "contour"
                        ]

        self.plotType = Struct()

        self.plotType.diag = Struct()
        self.plotType.diag.enabled = True
        self.plotType.diag.value = "histplot"
        self.plotType.diag.names = ["histplot"]

        self.plotType.upper = Struct()
        self.plotType.upper.enabled = True
        self.plotType.upper.value = "lineScatter"
        self.plotType.upper.names = dcopy(cornerNames)

        self.plotType.lower = Struct()
        self.plotType.lower.enabled = True
        self.plotType.lower.value = "contour"
        self.plotType.lower.names = dcopy(cornerNames)

        ############################################################################################################################
        #### setup subplot templates
        ############################################################################################################################

        template = Struct()

        # target

        template.target = Struct()
        template.target.axvline = Struct()
        template.target.axvline.kws = Struct()
        template.target.axhline = Struct()
        template.target.axhline.kws = Struct()
        template.target.scatter = Struct()
        template.target.scatter.kws = Struct()

        template.target.axvline.kws.linewidth = 0.5
        template.target.axvline.kws.linestyle = "-"
        template.target.axvline.kws.zorder = 1000
        template.target.axvline.kws.color = "orangered"

        template.target.axhline.kws.linewidth = 0.5
        template.target.axhline.kws.linestyle = "-"
        template.target.axhline.kws.zorder = 1000
        template.target.axhline.kws.color = "orangered"

        template.target.scatter.kws.s = 20
        template.target.scatter.kws.color = "orangered"
        template.target.scatter.kws.zorder = 1002

        # contour

        template.contour = Struct()
        template.contour.enabled = True
        template.contour.kws = Struct()
        template.contour.kws.cmap = "winter"
        template.contour.kws.alpha = 1
        template.contour.kws.levels = 10
        template.contour.kws.linewidths = 1
        template.contour.kws.linestyles = "solid"

        # contourf

        template.contourf = Struct()
        template.contourf.enabled = True
        template.contourf.kws = Struct()
        template.contourf.kws.cmap = "Blues"
        template.contourf.kws.alpha = 1
        template.contourf.kws.levels = 50

        # histplot

        template.histplot = Struct()
        template.histplot.enabled = True
        template.histplot.kws = Struct()
        template.histplot.kws.kde = False
        template.histplot.kws.bins = "auto"
        template.histplot.kws.stat = "count"
        template.histplot.kws.kde_kws = dict()
        template.histplot.kws.binwidth = None
        template.histplot.kws.binrange = None
        template.histplot.kws.multiple = "stack"
        template.histplot.kws.element = "step"
        template.histplot.kws.legend = False
        template.histplot.kws.shrink = 1
        template.histplot.kws.color = None
        template.histplot.kws.fill = True
        template.histplot.kws.common_norm = True
        template.histplot.kws.common_bins = False
        template.histplot.kws.line_kws = dict()
        template.histplot.kws.line_kws["linewidth"] = 0
        template.histplot.kws.line_kws["linestyle"] = "-"

        # legend template

        template.legend = Struct()
        template.legend.enabled = False
        template.legend.kws = Struct()

        # colorbar template

        template.colorbar = Struct()
        template.colorbar.enabled = False
        template.colorbar.kws = Struct()

        # line / scatter / lineScatter

        template.scatter = Struct()
        template.scatter.enabled = True
        template.scatter.kws = Struct()
        template.scatter.kws.s = 2
        template.scatter.kws.c = None
        template.scatter.kws.cmap = "winter"
        template.scatter.kws.alpha = 1
        template.scatter.kws.edgecolor = None
        template.scatter.kws.zorder = 2

        template.plot = Struct()
        template.plot.enabled = False
        template.plot.kws = Struct()
        template.plot.kws.linewidth = 1
        template.plot.kws.zorder = 1

        template.lineCollection = Struct()
        template.lineCollection.enabled = True
        template.lineCollection.kws = Struct()
        template.lineCollection.kws.cmap = "winter"
        template.lineCollection.kws.alpha = 1
        template.lineCollection.kws.linewidth = 1

        # figure

        template.figure = Struct()
        template.figure.kws = Struct()
        template.figure.enabled = False

        ############################################################################################################################
        #### setup subplots layout
        ############################################################################################################################

        self.layout = Struct()

        self.layout.contour                     = Struct()
        self.layout.contour.figure              = dcopy(template.figure)
        self.layout.contour.contour             = dcopy(template.contour)
        self.layout.contour.colorbar            = dcopy(template.colorbar)
        #self.layout.contour.target              = dcopy(template.target)
        self.layout.contour.noiseDensity        = 1.e-3
        self.layout.contour.limits              = None
        self.layout.contour.gridSize            = 512

        self.layout.contourf                    = Struct()
        self.layout.contourf.figure             = dcopy(template.figure)
        self.layout.contourf.contourf           = dcopy(template.contourf)
        self.layout.contourf.colorbar           = dcopy(template.colorbar)
        #self.layout.contourf.target             = dcopy(template.target)
        self.layout.contourf.noiseDensity       = 1.e-5
        self.layout.contourf.limits             = None
        self.layout.contourf.gridSize           = 512

        self.layout.histplot                    = Struct()
        self.layout.histplot.figure             = dcopy(template.figure)
        self.layout.histplot.histplot           = dcopy(template.histplot)
        self.layout.histplot.legend             = dcopy(template.legend)
        #self.layout.histplot.target             = dcopy(template.target)

        self.layout.line                        = Struct()
        self.layout.line.figure                 = dcopy(template.figure)
        self.layout.line.plot                   = dcopy(template.plot)
        self.layout.line.lineCollection         = dcopy(template.lineCollection)
        self.layout.line.colorbar               = dcopy(template.colorbar)
        self.layout.line.legend                 = dcopy(template.legend)
        #self.layout.line.target                 = dcopy(template.target)

        self.layout.scatter                     = Struct()
        self.layout.scatter.figure              = dcopy(template.figure)
        self.layout.scatter.scatter             = dcopy(template.scatter)
        self.layout.scatter.colorbar            = dcopy(template.colorbar)
        self.layout.scatter.legend              = dcopy(template.legend)
        #self.layout.scatter.target              = dcopy(template.target)

        self.layout.lineScatter                 = Struct()
        self.layout.lineScatter.figure          = dcopy(template.figure)
        self.layout.lineScatter.plot            = dcopy(template.plot)
        self.layout.lineScatter.scatter         = dcopy(template.scatter)
        self.layout.lineScatter.lineCollection  = dcopy(template.lineCollection)
        self.layout.lineScatter.colorbar        = dcopy(template.colorbar)
        self.layout.lineScatter.legend          = dcopy(template.legend)
        #self.layout.lineScatter.target          = dcopy(template.target)
        self.layout.lineScatter.plot.enabled    = True
        self.layout.lineScatter.plot.kws.alpha  = 0.2
        self.layout.lineScatter.plot.kws.color  = "grey"
        self.layout.lineScatter.plot.kws.linewidth = 0.75
        self.layout.lineScatter.lineCollection.enabled = False

        ############################################################################################################################

        self._isdryrun = True
        self.make()
        self._isdryrun = False

    ################################################################################################################################
    #### __call__
    ################################################################################################################################

    def __call__( self
                , reself : tp.Optional[ bool ] = False
                , **kwargs
                ):
        """

        Call the ``make()`` method of the current 
        instance of the class.

            **Parameters**

                Any arguments that can be passed to the 
                ``make()`` method of the plot object.

            **Returns**

                Any return value from the ``make()`` 
                method of the plot object.

        """

        return self.make(reself, **kwargs)

    ################################################################################################################################
    #### make
    ################################################################################################################################

    def make( self
            , reself : tp.Optional[ bool ] = False
            , **kwargs
            ):
        """

        Generate a grid plot from the selected 
        columns of the object's dataFrame.

            **Parameters**

                reself

                    A logical variable. If ``True``, an instance of
                    the object will be returned  to the calling routine
                    upon exit. The default value is ``False``.

            **Returns**

                The object self if ``reself = True`` otherwise, ``None``.
                However, this method causes side-effects by manipulating
                the existing attributes of the object.

        """

        for key in kwargs.keys():
            if hasattr(self,key):
                setattr(self, key, kwargs[key])
            elif key=="dataFrame":
                setattr( self, "_dfref", wref.ref(kwargs[key]) )
            else:
                raise Exception ( "Unrecognized input '"+key+"' class attribute detected." + newline
                                + self._getDocString()
                                )

        self._cEnabled = self.ccolumn is not None

        ############################################################################################################################
        #### verify the plot types to draw
        ############################################################################################################################

        if self.plotType.upper.enabled and (not (isinstance(self.plotType.upper.value,str) and self.plotType.upper.value in self.plotType.upper.names)):
            raise Exception ( "Unrecognized input value for the \"plotType.upper.value\" of the GridPlot object." + newline
                            + "The possible plot type names are given in \"plotType.upper.value\" of the object." + newline
                            + self._getDocString()
                            )

        if self.plotType.lower.enabled and (not (isinstance(self.plotType.lower.value,str) and self.plotType.lower.value in self.plotType.lower.names)):
            raise Exception ( "Unrecognized input value for the \"plotType.lower.value\" of the GridPlot object." + newline
                            + "The possible plot type names are given in \"plotType.lower.value\" of the object." + newline
                            + self._getDocString()
                            )

        if self.plotType.diag.enabled and (not (isinstance(self.plotType.diag.value,str) and self.plotType.diag.value in self.plotType.diag.names)):
            raise Exception ( "Unrecognized input value for the \"plotType.diag.value\" of the GridPlot object." + newline
                            + "The possible plot type names are given in \"plotType.diag.value\" of the object." + newline
                            + self._getDocString()
                            )

        ############################################################################################################################
        ############################################################################################################################
        if self._isdryrun: return
        ############################################################################################################################
        ############################################################################################################################

        import seaborn as sns
        import matplotlib.pyplot as plt

        ############################################################################################################################
        #### generate figure and axes if needed
        ############################################################################################################################

        self._constructBasePlot()

        ############################################################################################################################
        #### check data type
        ############################################################################################################################

        self._checkDataType()

        ############################################################################################################################
        #### check rows presence. This must be checked here, because it depends on the integrity of the in input dataFrame.
        ############################################################################################################################

        if self.rows is None: self.rows = range(len(self._dfref().index))

        ############################################################################################################################
        #### check columns presence. This must be checked here, because it depends on the integrity of the in input dataFrame.
        ############################################################################################################################

        # assign x columns to plot

        #if isinstance(self.columns,list):
        try:
            self._colnames, self._colindex = pm.dfutils.getColNamesIndex(self._dfref().columns,self.columns)
            colindexlen = len(self._colindex)
        except:
            raise Exception ( "The columns component of the current GridPlot object must point to" + newline
                            + "the names of a set of the columns of the input dataFrame to the GridPlot" + newline
                            + "class constructor." + newline
                            + self._getDocString()
                            )

        if colindexlen==0:
            raise Exception ( "The length of the ``columns`` component of the GridPlot object cannot be zero." + newline
                            + self._getDocString()
                            )

        # set color data

        if self._cEnabled:
            if isinstance(self.ccolumn,str) or isinstance(self.ccolumn,int):
                self._ccolname, self._ccolindex = pm.dfutils.getColNamesIndex(self._dfref().columns,self.ccolumn)
                #self._cdata = self._dfref().iloc[self.rows,self._colindex[0]].values.flatten()
            elif isinstance(self.ccolumn,list) and len(self.ccolumn)==0:
                self._ccolname = "Count"
                self._ccolindex = []
            else:
                raise Exception ( "The ccolumn component of the current GridPlot object must point to" + newline
                                + "the names of a column of the input dataFrame to the GridPlot" + newline
                                + "class constructor. It represents the set of values that will " + newline
                                + "be used to colormap the subplots, where needed." + newline
                                + self._getDocString()
                                )

            self._ccolindexlen = len(self._ccolindex)
            if self._ccolindexlen>1:
                raise Exception ( "The length of the ``ccolumn`` component of the GridPlot object cannot be larger than 1." + newline
                                + self._getDocString()
                                )


        ############################################################################################################################
        #### make pairgrid plot
        ############################################################################################################################

        self.currentFig.pairgrid = sns.PairGrid( self._dfref()[self._colnames].iloc[self.rows,:], **vars(self.pairgrid.kws) )
        self.currentFig.figure = self.currentFig.pairgrid.fig

        ############################################################################################################################
        #### hide axes as requested
        ############################################################################################################################

        if not self.plotType.upper.enabled: self.hide("upper")
        if not self.plotType.lower.enabled: self.hide("lower")
        if not self.plotType.diag.enabled: self.hide("diag")

        ############################################################################################################################
        #### adjust subplot interspaces
        ############################################################################################################################

        self.currentFig.pairgrid.fig.tight_layout()
        self.currentFig.pairgrid.fig.subplots_adjust(hspace = 0.05, wspace = 0.05)

        ############################################################################################################################
        #### add subplots
        ############################################################################################################################

        self._nrow = len(self.currentFig.pairgrid.axes[:])
        nrowMinusOne = self._nrow - 1
        self._ntot = self._nrow**2
        self._ntotStr = str(self._ntot)
        self.currentFig.subplotGrid = [ [ [] for j in range(self._nrow) ] for i in range(self._nrow) ]

        counter = 0
        dataMin = np.zeros(self._nrow)
        dataMax = np.zeros(self._nrow)
        dataMargin = np.zeros(self._nrow)

        ############################################################################################################################
        #### add the diagonal subplots
        ############################################################################################################################

        avgDiagTime = 0 # must be defined here for later default use, when diag is disabled.
        if self.plotType.diag.enabled:
            #self._progress.note( msg = "generating diagonal subplots... ", end = "" )
            self._progress.timer.toc()
            self.currentFig.pairgrid.map_diag( sns.histplot, **vars(self.layout.histplot.histplot.kws) )
            self._progress.timer.toc()
            avgDiagTime = self._progress.timer.delta / self._nrow
            #if self._reportEnabled:
            #    self._progress.timer.toc()
            #    print( "done in " + str(np.round(self._progress.timer.delta,6)) + " seconds." )

        ############################################################################################################################
        #### add the off-diagonal subplots
        ############################################################################################################################

        for irow, axRow in enumerate(self.currentFig.pairgrid.axes[:]):

            # compute data range

            #dataMin[irow] = np.min( self._dfref().iloc[self.rows,self._colindex[irow]].values.flatten() )
            #dataMax[irow] = np.max( self._dfref().iloc[self.rows,self._colindex[irow]].values.flatten() )
            #dataMargin[irow] = (dataMax[irow] - dataMin[irow]) / 10

            for icol, ax in enumerate(axRow):

                counter += 1
                plotEnabled = False

                self._progress.note( msg = "generating subplot #" + str(counter) + ": (" + str(irow) + "," + str(icol) + ") out of " + self._ntotStr + "... ", end = "" )

                if icol==irow and self.plotType.diag.enabled:
                
                    plotEnabled = True
                    requestedPlotType = self.plotType.diag.value
                    if self.plotType.diag.value=="histplot": RequestedPlotClass = DensityPlot

                if icol>irow and self.plotType.upper.enabled:

                    plotEnabled = True
                    requestedPlotType = self.plotType.upper.value
                    RequestedPlotClass = DensityPlot if "contour" in self.plotType.upper.value else LineScatterPlot

                if icol<irow and self.plotType.lower.enabled:

                    plotEnabled = True
                    requestedPlotType = self.plotType.lower.value
                    RequestedPlotClass = DensityPlot if "contour" in self.plotType.lower.value else LineScatterPlot

                if plotEnabled:

                    self.currentFig.subplotGrid[irow][icol] = RequestedPlotClass( plotType = requestedPlotType
                                                                                , dataFrame = self._dfref()
                                                                                , methodName = self._methodName
                                                                                , reportEnabled = False
                                                                                , resetPlot = None
                                                                                )

                    # set properties from the template

                    propsDist = dcopy( vars( getattr(self.layout,requestedPlotType) ) )
                    for key, val in propsDist.items(): setattr(self.currentFig.subplotGrid[irow][icol], key, val)

                    self.currentFig.subplotGrid[irow][icol].rows = self.rows
                    self.currentFig.subplotGrid[irow][icol].xcolumns = self._colnames[icol]

                    if not self.currentFig.subplotGrid[irow][icol]._type.is1d:

                        self.currentFig.subplotGrid[irow][icol].ycolumns = self._colnames[irow]

                        if "contourf"==self.currentFig.subplotGrid[irow][icol]._type.name:
                            if not self._cEnabled:
                                self.currentFig.subplotGrid[irow][icol].contourf.kws.cmap = None
                                # adding colors only makes a unicolor square subplot
                                #if "colors" not in vars(self.currentFig.subplotGrid[irow][icol].contourf.kws).keys():
                                #    self.currentFig.subplotGrid[irow][icol].contourf.kws.colors = "k"

                        elif "contour" in self.currentFig.subplotGrid[irow][icol]._type.name:
                            if not self._cEnabled:
                                self.currentFig.subplotGrid[irow][icol].contour.kws.cmap = None
                                #if "colors" not in vars(self.currentFig.subplotGrid[irow][icol].contour.kws).keys():
                                #if "colors" in vars(self.layput.contour.contour.kws).keys():
                                #    self.currentFig.subplotGrid[irow][icol].contour.kws.colors = "k"

                        else:

                            self.currentFig.subplotGrid[irow][icol].ccolumns = self.ccolumn

                    # make plot

                    plt.sca(ax)
                    if icol!=irow:
                        self.currentFig.subplotGrid[irow][icol].make()

                    # set the ticks and labels

                    if self.plotType.lower.enabled:
                        if icol > 0:
                            ax.set_ylabel("")
                        if irow < nrowMinusOne:
                            ax.set_xlabel("")
                    elif self.plotType.upper.enabled:
                        if icol > irow:
                            ax.set_ylabel("")
                            ax.set_xlabel("")

                # report progress

                if self._reportEnabled:
                    self._progress.timer.toc()
                    if irow==icol:
                        delta = avgDiagTime
                    else:
                        delta = self._progress.timer.delta
                    print( "done in " + str(np.round(delta,6)) + " seconds." )

        ############################################################################################################################

        plt.subplots_adjust(hspace=0.15, wspace=0.15)

        # add colorbar

        self._addColorBar()

        ############################################################################################################################

        if reself: return self

    ################################################################################################################################
    #### _addColorBar
    ################################################################################################################################

    def _addColorBar(self):

        if self._cEnabled and self.colorbar.enabled:

            self._progress.note( msg = "generating colorbar... ", end = "" )

            mappable = None
            isDensityMappable = False

            # search for line / scatter mappable

            if mappable is None and self.plotType.upper.enabled:

                if mappable is None and "scatter" in self.plotType.upper.value.lower():
                    mappable = self.currentFig.subplotGrid[0][1].currentFig.scatterList[0]

                if mappable is None and "line"==self.plotType.upper.value and self.layout.line.lineCollection.enabled:
                    mappable = self.currentFig.subplotGrid[0][1].currentFig.lineCollectionList[0]

                if mappable is None and "lineScatter"==self.plotType.upper.value and self.layout.lineScatter.lineCollection.enabled:
                    mappable = self.currentFig.subplotGrid[0][1].currentFig.lineCollectionList[0]

            if mappable is None and self.plotType.lower.enabled:

                if mappable is None and "scatter" in self.plotType.lower.value.lower():
                    mappable = self.currentFig.subplotGrid[1][0].currentFig.scatterList[0]

                if mappable is None and "line"==self.plotType.upper.value and self.layout.line.lineCollection.enabled:
                    mappable = self.currentFig.subplotGrid[1][0].currentFig.lineCollectionList[0]

                if mappable is None and "lineScatter"==self.plotType.upper.value and self.layout.lineScatter.lineCollection.enabled:
                    mappable = self.currentFig.subplotGrid[1][0].currentFig.lineCollectionList[0]

            # search for contour mappable

            if mappable is None and self.plotType.upper.enabled:

                if mappable is None and "contour"==self.plotType.upper.value.lower():
                    mappable = self.currentFig.subplotGrid[0][1].currentFig.contourList[0]
                    isDensityMappable = True

                if mappable is None and "contourf" in self.plotType.upper.value.lower():
                    mappable = self.currentFig.subplotGrid[0][1].currentFig.contourfList[0]
                    isDensityMappable = True

            if mappable is None and self.plotType.lower.enabled:

                if mappable is None and "contour"==self.plotType.lower.value.lower():
                    mappable = self.currentFig.subplotGrid[1][0].currentFig.contourList[0]
                    isDensityMappable = True

                if mappable is None and "contourf" in self.plotType.lower.value.lower():
                    mappable = self.currentFig.subplotGrid[1][0].currentFig.contourfList[0]
                    isDensityMappable = True

            # add colorbar

            if mappable is not None:

                self.colorbar.kws.mappable = mappable
                self.colorbar.kws.ax = self.currentFig.pairgrid.axes
                self.currentFig.colorbar = self.currentFig.figure.colorbar( **vars(self.colorbar.kws) )

                if isDensityMappable:
                    colorBarLabel = "Density"
                elif self._ccolindexlen==0:
                    colorBarLabel = "Count from the Series Start"
                else:
                    colorBarLabel = self._ccolname[0]
                self.currentFig.colorbar.set_label(colorBarLabel)

                self.currentFig.figure.set_figwidth(self.currentFig.figure.get_figwidth()*1.3333)

                self._progress.timer.toc()
                delta = self._progress.timer.delta
                msg = "done in " + str(np.round(delta,6)) + " seconds."

            else:

                msg = "No mappable were found in the current GridPlot. skipping..."

            if self._reportEnabled: print( msg, end = newline )

    ################################################################################################################################
    #### hide
    ################################################################################################################################

    def hide(self,part="all"):
        """

        Hides the requested part of the grid plot.

        **Parameters**

            part

                A string with the following possible values:

                    "lower"
                        hides the lower triangle of the grid plot.

                    "upper"
                        hides the upper triangle of the grid plot.

                    "diag"
                        hides the diagonal of the grid plot.

                    "all"
                        hides all grid plots and the colorbar.

                    "colorbar"
                        hides the colorbar of the grid plot.

                    The string can also be a mix of the above keywords,
                    separated by the ``+`` sign or some other delimiter.
                    For example, ``"lower+upper+colorbar"``

        **Returns**

            None

        """

        allhidden = "all" in part
        if allhidden or ("upper" in part): self.currentFig.pairgrid.map_upper(_hide_current_axis)
        if allhidden or ("lower" in part):
            self.currentFig.pairgrid.map_lower(_hide_current_axis)
            colnames = pm.dfutils.getColNamesIndex(self._dfref().columns,self.columns)[0]
            if not allhidden:
                for i in range(len(colnames)):
                    ax=self.currentFig.pairgrid.axes[:][i][i]
                    ax.set_xlabel(colnames[i])
                    ax.set_ylabel(colnames[i])
                    ax.xaxis.set_tick_params(which="both", labelbottom=True)
                    ax.yaxis.set_tick_params(which="both", labelbottom=True)
        if allhidden or ("diag" in part): self.currentFig.pairgrid.map_diag(_hide_current_axis)
        if allhidden or ("colorbar" in part): self.currentFig.colorbar.remove()
        #self.currentFig.figure.set_figwidth(self.currentFig.figure.get_figwidth()*1.3333)

    ################################################################################################################################
    #### show
    ################################################################################################################################

    def show(self,part="all"):
        """

        Shows the requested part of the grid plot.

            **Parameters**

                part
                    a string with the following possible values:

                        "lower"
                            shows the lower triangle of the grid plot.

                        "upper"
                            shows the lower triangle of the grid plot.

                        "diag"
                            shows the diagonal of the grid plot.

                        "all"
                            shows all grid plots.

                        "colorbar"
                            shows the colorbar of the grid plot.

                        The string can also be a mix of the above keywords,
                        separated by the ``+`` sign or some other delimiter.
                        For example, ``"lower+upper+colorbar"``

            **Returns**

                None

        """

        allshown = "all" in part
        if allshown or ("upper" in part): self.currentFig.pairgrid.map_upper(_show_current_axis)
        if allshown or ("lower" in part):
            self.currentFig.pairgrid.map_lower(_show_current_axis)
            colnames = pm.dfutils.getColNamesIndex(self._dfref().columns,self.columns)[0]
            colnamesLength=len(colnames)
            for i in range(colnamesLength):
                ax=self.currentFig.pairgrid.axes[:][i][i]
                if colnamesLength-1>i>0:
                    ax.set_xlabel("")
                    ax.set_ylabel("")
                    ax.xaxis.set_tick_params(which="both", labelbottom=False)
                    ax.yaxis.set_tick_params(which="both", labelbottom=False)
                elif i==0:
                    ax.xaxis.set_tick_params(which="both", labelbottom=False)
                    ax.set_xlabel("")
                else:
                    ax.yaxis.set_tick_params(which="both", labelbottom=False)
                    ax.set_ylabel("")
        if allshown or ("diag" in part): self.currentFig.pairgrid.map_diag(_show_current_axis)
        if allshown or ("colorbar" in part): self._addColorBar()

    ################################################################################################################################
    #### addTarget
    ################################################################################################################################

    def addTarget   ( self
                    , value : tp.Union[ np.ndarray , tp.List[float] , tp.Tuple[float] ] = None
                    ):
        """

        Call the target callable associated with each element of
        currentFig.subplotGrid of the GridPlot object to add target
        points and/or lines to all plots of the ``GridPlot`` figure.
        This method is supposed to be called only after a grid plot
        has been generated.

        **Parameters**

            value (optional)

                A numpy array, or list, or tuple whose length equals the
                number of columns/rows of the grid plot, each element of
                which represents the target value associated with the
                corresponding variable on the ``x`` axis of the plot,
                from the left to the right.

                Alternatively, ``value`` can be a string with the following 
                possible value:

                    "mode"

                        The variable values corresponding to the mode of the 
                        "SampleLogFunc" column of the input dataFrame will 
                        be used. If no "SampleLogFunc" columns name exists 
                        in the input dataFrame, an exception will be raised.

                        This is the **default value** for the input variable 
                        ``value``.

                    "mean"

                        The variable values corresponding to the mode of the 
                        "SampleLogFunc" column of the input dataFrame will 
                        be used. If no "SampleLogFunc" columns name exists 
                        in the input dataFrame, an exception will be raised.

                If not provided, the default will be set to the state
                (coordinates) of the mode of the "SampleLogFunc" column 
                in the input dataFrame to the object. This would work only
                when the input dataFrame is the contents of a ParaMonte
                output chain or sample file.

        **Returns**

            None. However, this method causes side-effects by manipulating
            the existing attributes of the ``Target`` objects in 
            ``currentFig.subplotGrid`` of the GridPlot object.

        """

        if value is None: value="mode"

        if isinstance(value, str):
            if value=="mode":
                from _dfutils import getMaxLogFunc
                maxLogFunc = getMaxLogFunc(dataFrame=self._dfref().iloc[self.rows,:])
                value = maxLogFunc.dfrow[self._colindex]
            elif value=="mean":
                value = self._dfref()[self._colnames].iloc[self.rows,:].mean()
            elif value=="median":
                value = self._dfref()[self._colnames].iloc[self.rows,:].median()
        else:
            from collections.abc import Iterable
            if not isinstance(value, Iterable):
                raise Exception ( newline
                                + "The input argument ``value`` must be either a string or an array of target value." + newline
                                + "Here is the help information for addTarget():"
                                + newline
                                + self.addTarget.__doc__
                                )

        counter = 0
        for irow, plotObjectRow in enumerate(self.currentFig.subplotGrid[:]):

            for icol, plotObject in enumerate(plotObjectRow):

                counter += 0
                self._progress.note( msg = "generating target #" + str(counter) + ": (" + str(irow) + "," + str(icol) + ") out of " + self._ntotStr + "... ", end = "" )

                #import matplotlib.pyplot as plt
                #plt.sca(self.currentFig.pairgrid.axes[irow][icol])
                if icol>irow and self.plotType.upper.enabled:
                    plotObject.target( value = [ value[icol], value[irow] ], axes = self.currentFig.pairgrid.axes[irow][icol] )
                elif icol<irow and self.plotType.lower.enabled:
                    plotObject.target( value = [ value[icol], value[irow] ], axes = self.currentFig.pairgrid.axes[irow][icol] )
                elif icol==irow and self.plotType.diag.enabled:
                    plotObject.target( value = [ value[icol], value[irow] ], axes = self.currentFig.pairgrid.axes[irow][icol] )

                # report progress

                if self._reportEnabled:
                    self._progress.timer.toc()
                    delta = self._progress.timer.delta
                    print( "done in " + str(np.round(delta,6)) + " seconds.", end = newline )

    ################################################################################################################################
    #### _getDocString
    ################################################################################################################################

    def _getDocString(self):
        docString   = newline \
                    + "Here is the help information on the GridPlot class:" + newline \
                    + newline \
                    + self.__doc__ \
                    + super()._getDocString()
        return docString

    ################################################################################################################################
    #### helpme
    ################################################################################################################################

    def helpme(self, topic=None):
        """

        Print the documentation for the input string topic. 
        If the topic does not exist, the documentation for
        the object will be printed.
     
            **Parameters**
         
                topic (optional)

                A string containing the name of the object 
                for which help is needed.

            **Returns**
         
                None
         
            **Example**

                .. code-block:: python
                    :linenos:

                    helpme()
                    helpme("make")
                    helpme("helpme")
                    helpme("getLogLinSpace")

        """

        try:
            exec("print(self."+topic+".__doc__)")
        except:
            print(self._getDocString())
        return None

    ################################################################################################################################

####################################################################################################################################

def _hide_current_axis(*args, **kwds):
    from matplotlib import pyplot as plt
    plt.gca().set_visible(False)

####################################################################################################################################

def _show_current_axis(*args, **kwds):
    from matplotlib import pyplot as plt
    plt.gca().set_visible(True)

####################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a
####   copy of this software and associated documentation files (the "Software"),
####   to deal in the Software without restriction, including without limitation
####   the rights to use, copy, modify, merge, publish, distribute, sublicense,
####   and/or sell copies of the Software, and to permit persons to whom the
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your
####   work (education/research/industry/development/...) by citing the ParaMonte
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np

####################################################################################################################################
#### colorbar
####################################################################################################################################

def isColorBar(ax):
    """

    Attempt to guess whether an input matplotlib
    ``Axes`` is home to a ``colorbar`` object.

        **Parameters**

            ax

                An Axes instance.

        **Returns**

            A boolean value of ``True`` if the x xor y axis
            satisfies all of the following and thus looks
            like a ``colorbar`` object:
                No ticks, no tick labels, no axis label

    """

    xcb = (len(ax.get_xticks()) == 0) and (len(ax.get_xticklabels()) == 0) and (len(ax.get_xlabel()) == 0) #and (ax.get_xlim() == (0, 1))
    ycb = (len(ax.get_yticks()) == 0) and (len(ax.get_yticklabels()) == 0) and (len(ax.get_ylabel()) == 0) #and (ax.get_ylim() == (0, 1))
    return xcb != ycb

####################################################################################################################################
#### isPresentColorBar
####################################################################################################################################

def isPresentColorBar():
    from matplotlib import pyplot as plt
    return any([isColorBar(ax) for ax in np.atleast_1d(plt.gcf().axes).flatten()])



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import setuptools

with open("README.md", "r") as fh: readmeFileContents = fh.read()

with open("paramonte/auxil/.VERSION_INTERFACE") as versionFile: version = versionFile.readline()

setuptools.setup( name                          = "paramonte"
                , version                       = version
                , author                        = "Amir Shahmoradi, Fatemeh Bagheri, Joshua Alexander Osborne"
                , author_email                  = "shahmoradi@utexas.edu"
                , description                   = "Plain Powerful Parallel Monte Carlo and adaptive MCMC Library"
                , long_description              = readmeFileContents
                , long_description_content_type = "text/markdown"
                , url                           = "https://github.com/cdslaborg/paramonte"
                , packages                      = setuptools.find_packages()
                , python_requires               = '>=3.0'
                , license                       = "License :: OSI Approved :: MIT License"
                , include_package_data          = True
                , install_requires              =   [ "numpy"
                                                    , "pandas"
                                                    #, "scipy"
                                                    #, "seaborn"
                                                    #, "matplotlib"
                                                    #, OTHERS:
                                                    #, "platform"
                                                    #, "weakref"
                                                    #, "typing"
                                                    #, "time"
                                                    #, "sys"
                                                    #, "os"
                                                    ]
                , classifiers                   =   [ "Development Status :: 5 - Production/Stable"
                                                    , "Natural Language :: English"
                                                    , "Programming Language :: Python :: 3"
                                                    , "Programming Language :: Fortran"
                                                    , "Programming Language :: C"
                                                    , "License :: OSI Approved :: MIT License"
                                                    , "Operating System :: OS Independent"
                                                    , "Topic :: Scientific/Engineering"
                                                    , "Topic :: Scientific/Engineering :: Physics"
                                                    , "Topic :: Scientific/Engineering :: Astronomy"
                                                    , "Topic :: Scientific/Engineering :: Mathematics"
                                                    , "Topic :: Scientific/Engineering :: Visualization"
                                                    , "Topic :: Scientific/Engineering :: Bio-Informatics"
                                                    , "Topic :: Scientific/Engineering :: Atmospheric Science"
                                                    , "Operating System :: OS Independent"
                                                    , "Operating System :: Microsoft :: Windows"
                                                    , "Operating System :: POSIX :: Linux"
                                                    , "Operating System :: MacOS"
                                                    , "Operating System :: Unix"
                                                    ]
                )



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import os
import sys
import platform as _platform

import _dfutils as dfutils
import _pmutils as utils
from _message import note, warn, abort
from _pmutils import Struct, newline, creturn

####################################################################################################################################

from pathlib import Path as _Path

path = Struct()
path.root = os.path.dirname(os.path.abspath(__file__))
path.auxil = os.path.join(path.root,"auxil")
path.home = str(_Path.home()) # path.home = os.path.expanduser("~")
path.lib = path.root

sys.path.append(path.root)

####################################################################################################################################

names = Struct()
names.paramonte = "ParaMonte"
names.paradram = "ParaDRAM"
names.paradise = "ParaDISE"
names.paranest = "ParaNest"
names.paratemp = "ParaTemp"

####################################################################################################################################

platform = Struct()
platform.arch = "x86" if "32" in _platform.architecture()[0] else "x64"
platform.name = sys.platform.lower()
platform.isWin32 = True if platform.name=="win32" else False
platform.isLinux = True if platform.name=="linux" else False
platform.isMacOS = True if platform.name=="darwin" else False
platform.osname = "windows" if platform.isWin32 else platform.name

from datetime import datetime as _dt
dt = _dt.now()
platform.systemInfoFilePrefix = os.path.join(path.auxil, ".systemInfo_");
platform.systemInfoFilePath = platform.systemInfoFilePrefix + "{:04d}".format(dt.year) + "{:02d}".format(dt.month) + "{:02d}".format(dt.day);

if not os.path.isfile(platform.systemInfoFilePath):

    import glob
    fileList = glob.glob(platform.systemInfoFilePrefix + "*")
    for filePath in fileList:
        try:
            os.remove(filePath)
        except:
            pass

    cmd = None
    if platform.isWin32: cmd = "systeminfo > " + platform.systemInfoFilePath
    if platform.isLinux: cmd = "uname -a >>  " + platform.systemInfoFilePath + "; lscpu >> " + platform.systemInfoFilePath
    if platform.isMacOS: cmd = "uname -a >>  " + platform.systemInfoFilePath + "; sysctl -a | grep machdep.cpu >> " + platform.systemInfoFilePath

    if cmd is not None:
        err = os.system(cmd)
        if err != 0:
            platform.systemInfoFilePath = None
            warn( msg   = "Failed to get the system information. skipping..."
                , methodName = names.paramonte
                , marginTop = 1
                , marginBot = 1
                )

####################################################################################################################################

website = Struct()

website.home = Struct()
website.home.url = "https://www.cdslab.org/paramonte"
website.home.install = Struct()
website.home.install._url = website.home.url + "/notes/installation"

# installation Linux

website.home.overview = Struct()
website.home.overview._url = website.home.url + "/notes/overview"
website.home.overview.preface = Struct()
website.home.overview.changes = Struct()
website.home.overview.preface.url = website.home.overview._url + "/preface"
website.home.overview.changes.kernel = Struct()
website.home.overview.changes.matlab = Struct()
website.home.overview.changes.python = Struct()
website.home.overview.changes.kernel.url = website.home.overview._url + "/paramonte-kernel-release-notes"
website.home.overview.changes.matlab.url = website.home.overview._url + "/paramonte-matlab-release-notes"
website.home.overview.changes.python.url = website.home.overview._url + "/paramonte-python-release-notes"

# installation Linux

website.home.install.linux = Struct()
website.home.install.linux.url = website.home.install._url + "/linux"

# installation Windows

website.home.install.windows = Struct()
website.home.install.windows.url = website.home.install._url + "/windows"

# installation Python

website.home.install.python = Struct()
website.home.install.python.url = website.home.install._url + "/python"

# installation MATLAB

website.home.install.matlab = Struct()
website.home.install.matlab.url = website.home.install._url + "/matlab"

# installation macOS

website.home.install.macos = Struct()
website.home.install.macos.url = website.home.install._url + "/macos"
website.home.install.macos.prereqs = Struct()
website.home.install.macos.prereqs.url = website.home.install.macos.url + "/#the-compile-time-and-runtime-prerequisites"
website.home.install.macos.prereqs.cmd = Struct()
website.home.install.macos.prereqs.cmd.url = website.home.install.macos.url + "/#prereqs-install"

# Python examples

website.home.examples = Struct()
website.home.examples._url = website.home.url + "/notes/examples"
website.home.examples.python = Struct()
website.home.examples.python.jupyter = Struct()
website.home.examples.python.postprocess = Struct()
website.home.examples.python.jupyter.url = website.home.examples._url + "/python/jupyter"
website.home.examples.python.postprocess.url = website.home.examples._url + "/python/postprocess"

# Python API

website.home.api = Struct()
website.home.api._url = website.home.url + "/notes/api"
website.home.api.python = Struct()
website.home.api.python.url = website.home.api._url + "/python/autoapi/paramonte"

# ParaDRAM

website.home.usage = Struct()
website.home.usage._url = website.home.url + "/notes/usage"
website.home.usage.paradram = Struct()
website.home.usage.paradram._url = website.home.usage._url + "/paradram"
website.home.usage.paradram.quickstart = Struct()
website.home.usage.paradram.quickstart.url = website.home.usage.paradram._url + "/interface"
website.home.usage.paradram.input = Struct()
website.home.usage.paradram.input.url = website.home.usage.paradram._url + "/input"
website.home.usage.paradram.specifications = Struct()
website.home.usage.paradram.specifications.url = website.home.usage.paradram._url + "/specifications"
website.home.usage.paradram.restart = Struct()
website.home.usage.paradram.restart.url = website.home.usage.paradram._url + "/restart"
website.home.usage.paradram.output = Struct()
website.home.usage.paradram.output.url = website.home.usage.paradram._url + "/output"

# GitHub issues

website.github = Struct()
website.github.url = "https://github.com/cdslaborg/paramonte"
website.github.issues = Struct()
website.github.issues.url = "https://github.com/cdslaborg/paramonte/issues"
website.github.release = Struct()
website.github.release.url = website.github.url + "/releases"
website.github.release.latest = Struct()
website.github.release.latest.url = website.github.release.url + "/latest"
website.github.archive = Struct()
website.github.archive._url = website.github.url + "/archive"
website.github.archive.master = Struct()
website.github.archive.master.zip = Struct()
website.github.archive.master.tar = Struct()
website.github.archive.master.zip.url = website.github.archive._url + "/master.zip"
website.github.archive.master.tar.url = website.github.archive._url + "/master.tar.gz"

# GitHub examples

website.github.examples = Struct()
website.github.examples.url = "https://github.com/cdslaborg/paramontex"

# Intel MPI

website.intel = Struct()
website.intel.mpi = Struct()
website.intel.mpi.home = Struct()
website.intel.mpi.home.url = "https://software.intel.com/en-us/mpi-library"

# Intel Windows

website.intel.mpi.windows = Struct()
website.intel.mpi.windows.url = "https://software.intel.com/en-us/get-started-with-mpi-for-windows"

# OpenMPI

website.openmpi = Struct()
website.openmpi.home = Struct()
website.openmpi.home.url = "https://www.open-mpi.org/"

####################################################################################################################################

def cite(): print(website.home.overview.preface.url + "/#how-to-acknowledge-the-use-of-the-paramonte-library-in-your-work")
#citation.kernel = Struct()
#citation.kernel.paradram = Struct()
#citation.kernel.paradram.bib = """
#@article{2020arXiv200809589S,
#           author = {{Shahmoradi}, Amir and {Bagheri}, Fatemeh},
#            title = "{ParaDRAM: A Cross-Language Toolbox for Parallel High-Performance Delayed-Rejection Adaptive Metropolis Markov Chain Monte Carlo Simulations}",
#          journal = {arXiv e-prints},
#         keywords = {Computer Science - Computational Engineering, Finance, and Science, Astrophysics - Instrumentation and Methods for Astrophysics, Physics - Data Analysis, Statistics and Probability, Statistics - Computation, Statistics - Machine Learning},
#             year = 2020,
#            month = aug,
#              eid = {arXiv:2008.09589},
#            pages = {arXiv:2008.09589},
#    archivePrefix = {arXiv},
#           eprint = {2008.09589},
#     primaryClass = {cs.CE},
#           adsurl = {https://ui.adsabs.harvard.edu/abs/2020arXiv200809589S},
#          adsnote = {Provided by the SAO/NASA Astrophysics Data System}
#}
#"""

####################################################################################################################################

from _Version import Version
version = Struct()
for versionType in ["interface","kernel"]: setattr(version,versionType,Version(path.auxil,versionType))

####################################################################################################################################




####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
import typing as tp
import pandas as pd
import weakref as wref

import _message as err
import _dfutils as dfutils
import _pmutils as pmutils
from paramonte.vis.LineScatterPlot import LineScatterPlot

Struct = pmutils.Struct
newline = pmutils.newline

####################################################################################################################################
#### AutoCorr class
####################################################################################################################################

class AutoCorr:
    """

    This is the base class for generating object of type ``AutoCorr``.  
    Upon construction, it will provide methods to compute and plot the 
    autocorrelations of the selected columns of the input dataFrame.

        **Parameters**

            dataFrame

                A Pandas dataframe based upon the selected columns 
                of which the autocorrelations will be computed.

            columns (optional)

                A list that determines the columns of the input dataFrame 
                to be used in the computation of the autocorrelations. 
                It can have three forms:

                    1.  A list of column indices from the input dataFrame.
                    2.  A list of column names from dataFrame.columns.
                    3.  A ``range(start,stop,step)`` of column indices.

                Examples:

                    1.  ``columns = [0,1,4,3]``
                    2.  ``columns = ["SampleLogFunc","SampleVariable1"]``
                    3.  ``columns = range(17,7,-2)``

                The default behavior includes all columns of the dataFrame.

        **Attributes**

            All of the parameters described above, except dataFrame.

                A weak reference to the dataFrame will be implicitly 
                stored in the object. 

            df

                A pandas dataframe containing the computed autocorrelations.

            rows

                A list of the dataFrame indices that determines the rows of 
                the input dataFrame to be used in the computation of the 
                autocorrelations. It can be either:

                    1.  A ``range(start,stop,step)``, or, 
                    2.  A list of row indices from ``dataFrame.index``.

                Examples:

                    1.  ``rows = range(17,7,-2)``
                    2.  ``rows = [i for i in range(7,17)]``

                The default behavior includes all rows of the dataFrame.

            plot

                A structure containing the following plotting tools:

                    heatmap

                        A callable object of class HeatMap which will 
                        enable plotting of the correlation matrix.

        **Returns**

            self

                an object of type class CorMat.

    ---------------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , dataFrame     # : tp.Optional[ pd.DataFrame ]
                , columns       = None # : tp.Optional[ tp.Union[ range , tp.List[int] , tp.List[str] ] ] = None
                , methodName    = "ParaMonte" # : tp.Optional[ str ] = "ParaMonte"
                , reportEnabled = True # : tp.Optional[ bool ] = True
                ):

        self.df = None
        self.rows = None
        self.columns = columns

        self._dfref = None if dataFrame is None else wref.ref(dataFrame)

        self._methodName = methodName
        self._reportEnabled = reportEnabled
        self._progress = pmutils.Progress   ( msg = None
                                            , methodName = methodName
                                            , reportEnabled = reportEnabled
                                            , end = "\n"
                                            )

    ################################################################################################################################
    #### __call__
    ################################################################################################################################

    def __call__( self
                , reself : tp.Optional[ bool ] = False
                , **kwargs
                ):
        """

        Call the ``get()`` method of the current instance of the class.

            **Parameters**

                Any arguments that can be passed to the ``get()`` 
                method of the object.

            **Returns**

                Any return value from the ``get()`` method of the object.

        """

        return self.get(reself, **kwargs)

    ################################################################################################################################
    #### get
    ################################################################################################################################

    def get ( self
            , reself : tp.Optional[ bool ] = False
            , **kwargs
            ):
        """

        Compute the autocorrelations of the selected columns 
        of the input dataframe to the object's constructor.

            **Parameters**

                reself

                    A logical variable. When ``True``, an instance of 
                    the object will be returned to the calling routine 
                    upon exit. The default value is ``False``.

            **Returns**

                The object if ``reself = True`` otherwise, ``None``.

                **NOTE**

                This method causes side-effects by manipulating
                the existing attributes of the object.

        """

        for key in kwargs.keys():
            if hasattr(self,key):
                setattr(self, key, kwargs[key])
            elif key=="dataFrame":
                setattr( self, "_dfref", wref.ref(kwargs[key]) )
            else:
                raise Exception ( "Unrecognized input '"+key+"' class attribute detected." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        #### check columns presence
        ############################################################################################################################

        colnames, colindex = dfutils.getColNamesIndex(self._dfref().columns,self.columns)

        ############################################################################################################################
        #### check rows presence. This must be checked here, because it depends on the integrity of the in input dataFrame.
        ############################################################################################################################

        if self.rows is None: self.rows = range(len(self._dfref().index))
        #rownames = self._dfref().index[self.rows]

        ############################################################################################################################
        #### compute the autocorrelations
        ############################################################################################################################

        self._nvar = len(colnames)
        self._nlag = len(self.rows)
        acf = np.zeros((self._nvar,self._nlag))

        try:

            from scipy.signal import correlate
            for cnt, ivar in enumerate(colindex):
                xdata = self._dfref().iloc[self.rows,ivar].values.flatten() - np.mean(self._dfref().iloc[self.rows,ivar].values.flatten())
                acf[cnt] = correlate( xdata
                                    , xdata
                                    , mode = "full"
                                    )[self._nlag-1:2*self._nlag]
                acf[cnt] = acf[cnt] / acf[cnt,0]

        except:

            if self._reportEnabled:
                err.warn( msg   = "Failed to compute the Autocorrelation function of the input dataFrame." + newline
                                + "This could likely be due to an error in importing the scipy Python library." + newline
                                + "Please make sure you have the scipy library properly installed on your system." + newline
                                + "You can do so by typing the following command on your Anaconda3 or Bash command prompt:" + newline
                                + newline
                                + "    pip install --user --upgrade scipy"
                        , methodName = self._methodName
                        , marginTop = 1
                        , marginBot = 1
                        )

        self.df = pd.DataFrame(np.transpose(acf))

        ############################################################################################################################
        #### specify columns/index names
        ############################################################################################################################

        colnames = [ "ACF_"+colnames[i] for i in range(len(colnames)) ]
        self.df.columns = colnames

        ############################################################################################################################
        #### add SampleLogFunc to df for plot coloring, if exists
        ############################################################################################################################

        ccolumns = ()
        if "SampleLogFunc" in self._dfref().columns:
            ccolumns = "SampleLogFunc"
            self.df.insert  ( loc = 0
                            , column = ccolumns
                            , value = self._dfref()[[ccolumns]].values.flatten()
                            , allow_duplicates = True
                            )

        ############################################################################################################################
        #### add lags to df
        ############################################################################################################################

        self.df.insert  ( loc = 0
                        , column = "Lag"
                        , value = [ i for i in self.df.index ]
                        )

        ############################################################################################################################
        #### graphics
        ############################################################################################################################

        self._plotTypeList =    [ "line"
                                , "scatter"
                                , "lineScatter"
                                ]

        self._progress.note( msg = "adding the autocrrelation graphics tools... ", end = newline, pre = True )
        self.plot = Struct()
        self._resetPlot(resetType="hard")

        self.plot.reset = self._resetPlot

        ############################################################################################################################

        if reself: return self

    ################################################################################################################################
    #### _resetPlot
    ################################################################################################################################

    def _resetPlot  ( self
                    , resetType = "soft"
                    , plotNames = "all"
                    ):
        """

        Reset the properties of the plot to the original default settings.
        Use this method when you change many attributes of the plot and
        you want to clean up and go back to the default settings.

            **Parameters**

                resetType (optional)

                    An optional string with possible value of ``"hard"``.
                    If provided, the plot object will be regenerated from scratch.
                    This includes reading the original data frame again and resetting
                    everything. If not provided, then only the plot settings will be
                    reset without reseting the dataFrame.

                plotNames (optional)

                    An optional string value or list of string values representing 
                    the names of plots to reset. If no value is provided, 
                    then all plots will be reset.

            **Returns**

                None

            **Example**

                .. code-block:: python

                    reset("hard")                    # regenerate all plots from scratch
                    reset("hard","line")             # regenerate line plot from scratch
                    reset("hard",["line","scatter"]) # regenerate line and scatter plots

        """

        if isinstance(plotNames, str):
            plotTypeLower = plotNames.lower()
            if plotTypeLower=="all":
                requestedPlotTypeList = self._plotTypeList
            elif plotNames in self._plotTypeList:
                requestedPlotTypeList = [plotNames]
            else:
                self._reportWrongPlotName(plotNames)
        elif isinstance(plotNames, list):
            for plotName in plotNames:
                if plotName not in self._plotTypeList: self._reportWrongPlotName(plotName)
        else:
            self._reportWrongPlotName("a none-string none-list object.")

        resetTypeIsHard = None
        if isinstance(resetType, str):
            resetTypeIsHard = resetType.lower()=="hard"
        else:
            err.abort   ( msg   = "The input argument resetType must be a string representing" + newline
                                + "the type of the reset to be performed on the plots." + newline
                                + "A list of possible plots includes: \"hard\", \"soft\"" + newline
                                + "Here is the help for the ``reset()`` method: " + newline
                                + newline
                                + self._resetPlot.__doc__
                        , marginTop = 1
                        , marginBot = 1
                        , methodName = self._methodName
                        )

        ############################################################################################################################
        #### reset plots
        ############################################################################################################################

        for requestedPlotType in requestedPlotTypeList:

            plotObject = None
            requestedPlotTypeLower = requestedPlotType.lower()

            isLine      = "line"        in requestedPlotTypeLower
            isScatter   = "scatter"     in requestedPlotTypeLower

            isLineScatterPlot = isLine or isScatter

            if not resetTypeIsHard:
                plotComponent = getattr(self, "plot")
                plotObject = getattr(plotComponent, requestedPlotType)
                plotObject._reset()

            ########################################################################################################################
            #### reset line / scatter
            ########################################################################################################################

            if isLineScatterPlot:

                if resetTypeIsHard:
                    plotObject = LineScatterPlot( plotType = requestedPlotType
                                                , dataFrame = self.df
                                                , methodName = self._methodName
                                                , reportEnabled = self._reportEnabled
                                                , resetPlot = self._resetPlot
                                                )

                plotObject.xcolumns = "Lag"
                plotObject.ycolumns = self.df.columns[2:]

                plotObject._xlabel = "AutoCorrelation Lag"
                plotObject._ylabel = "AutoCorrelation Function (ACF) Value"

                plotObject.ccolumns = None

                plotObject.colorbar.kws.extend = "neither"
                plotObject.colorbar.kws.orientation = "vertical"
                plotObject.colorbar.kws.spacing = "uniform"

                plotObject._xlimit = [1, None]
                plotObject._ylimit = [None, 1]
                plotObject._xscale = "log"

                plotObject.legend.enabled = True

                if isLine:
                    if isScatter:
                        plotObject.lineCollection.enabled = False
                        plotObject.plot.enabled = True
                        plotObject.plot.kws.alpha = 0.2
                        plotObject.plot.kws.color = "grey"
                        plotObject.plot.kws.linewidth = 0.75
                    else:
                        plotObject.plot.enabled = True
                        plotObject.plot.kws.linewidth = 1
                        plotObject.lineCollection.enabled = False

            ########################################################################################################################

            if plotObject is not None: setattr(self.plot, requestedPlotType, plotObject)

    ################################################################################################################################
    #### _reportWrongPlotName
    ################################################################################################################################

    def _reportWrongPlotName( self
                            , plotNames
                            ):

        err.abort   ( msg   = "The input argument plotNames must be a string representing" + newline
                            + "the name of a plot belonging to the AutoCorr class or," + newline
                            + "a list of such plot names. You have entered: " + plotNames + newline
                            + "Possible plots are: " + newline
                            + newline
                            + newline.join(self._plotTypeList) + newline
                            + newline
                            + "Here is the help for the ``reset()`` method: " + newline
                            + newline
                            + self._resetPlot.__doc__
                    , marginTop = 1
                    , marginBot = 1
                    , methodName = self._methodName
                    )

    ################################################################################################################################




#!/usr/bin/env python
#!C:\ProgramData\Anaconda3\python.exe
####################################################################################################################################
##  
##  Description:
##      +   Run the Monte Carlo sampler of the ParaMonte library given the input log-target density function `getLogFunc()`.
##  Output:
##      +   The simulation output files.
##  Author:
##      +   Computational Data Science Lab, Monday 9:03 AM, May 16 2016, ICES, UT Austin
##  Visit:
##      +   https://www.cdslab.org/paramonte
##
####################################################################################################################################

import os
import numpy as np
import paramonte as pm
from logfunc import getLogFunc, NDIM

pmpd = pm.ParaDRAM() # define a ParaDRAM sampler instance

pmpd.mpiEnabled = True # flag to request a parallel simulation

pmpd.runSampler ( ndim = NDIM               # number of dimensions of the objective function
                , getLogFunc = getLogFunc   # the objective function: multivariate normal distribution
                # NOTE: inputFile is optional: all simulation specifications can be set as attributes of pmpd.spec
                # NOTE: However, if specified, then all simulation specification will be read exclusively from the inputFile.
                , inputFile = os.path.dirname(os.path.abspath(__file__)) + "/paramonte.in"
                )



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a
####   copy of this software and associated documentation files (the "Software"),
####   to deal in the Software without restriction, including without limitation
####   the rights to use, copy, modify, merge, publish, distribute, sublicense,
####   and/or sell copies of the Software, and to permit persons to whom the
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your
####   work (education/research/industry/development/...) by citing the ParaMonte
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
import _paramonte as pm
import _CorCovMat as ccm
from _AutoCorr import AutoCorr
from _dfutils import getMaxLogFunc
from _OutputFileContents import OutputFileContents
from paramonte.vis.LineScatterPlot import LineScatterPlot
from paramonte.vis.DensityPlot import DensityPlot
from paramonte.vis.GridPlot import GridPlot

Struct = pm.Struct
newline = pm.newline

####################################################################################################################################
#### TabularFileContents class
####################################################################################################################################

class TabularFileContents(OutputFileContents):
    """

    This is the **TabularFileContents** class for generating instances
    of the ParaMonte tabular output contents. This class is NOT meant to
    be directly accessed by the ParaMonte library users. It is internally
    used by the ParaMonte library to parse the tabular contents of the
    output files generated by the ParaMonte sampler routines. For example,
    the ParaDRAM sampler class makes calls to this class via its
    ``readSample()`` or ``readChain()`` or ``readMarkovChain()``
    or ``readProgress()`` methods to return a list of objects
    of class ``TabularFileContents``.

        **Parameters**

            file

                The full path to the file containing the sample/chain.

            fileType

                A string containing the type of the file to be parsed.
                Current options include but are not limited to:
                ``sample``, ``chain``, ``markovChain``, ``progress``

            delimiter

                The delimiter used in the sample/chain file, which
                must be provided by the user.

            methodName

                A string representing the name of the ParaMonte sampler used
                to call the constructor of the ``TabularFileContents`` class.

            parseContents

                If set to ``True``, the contents of the file will be parsed and
                stored in a component of the object named ``contents``.
                The default value is ``True``.

            reportEnabled

                A logical input parameter indicating whether the ParaMonte
                automatic guidelines to the standard output should be provided
                or not. The default value is ``True``.

        **Attributes**

            file

                The full path to the file containing the sample/chain.

            delimiter

                The delimiter used in the sample/chain file, which
                must be provided by the user.

            ndim

                The number of dimensions of the domain of the objective
                function from which the sample has been drawn.

            count

                The number of points (states) in the sample/chain file.
                This is essentially, the number of rows in the file
                minus one (representing the header line).

            plot

                A structure containing the graphics tools for the
                visualization of the contents of the file.

            df

                If the input file contents is structured in a format that
                could be read as a dataframe, then the contents of the file
                will be stored in the form of a pandas-library DataFrame
                in this property (hence called ``df``).

            contents

                If the input file contents is structured in the form of columns,
                then a property named ``contents`` is also added to the object.
                Each component of contents will named via the header of the file
                and will contain data from the corresponding column of the file.

        **Returns**

            tabularFileContents

                An object of class ``TabularFileContents``.

    ----------------------------------------------------------------------
    """

    def __init__( self
                , file
                , fileType
                , delimiter
                , methodName
                , parseContents = True
                , reportEnabled = True
                ):

        super().__init__(file, methodName, reportEnabled)


        markovChainRequested = fileType=="markovChain"
        self._isProgressFile = "progress"==fileType

        self._sampleLogFuncColName = "" if self._isProgressFile else "SampleLogFunc"

        #if "sample"==fileType:
        #    fileSuffix = "sample"
        #elif fileType=="chain" or markovChainRequested:
        #    fileSuffix = "chain"
        #elif self._isProgressFile:
        #    fileSuffix = "progress"
        #else:
        #    pm.abort( msg   = "Internal error occurred. The input fileType is not recognized.\n"
        #                    + "Please report this error at:\n\n"
        #                    + "    " + pm.website.github.issues.url
        #            , methodName = self._methodName
        #            , marginTop = 1
        #            , marginBot = 1
        #            )

        if fileType!="sample" and fileType!="chain" and not (self._isProgressFile or markovChainRequested):
            pm.abort( msg   = "Internal error occurred. The input fileType is not recognized.\n"
                            + "Please report this error at:\n\n"
                            + "    " + pm.website.github.issues.url
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        ############################################################################################################################
        #### data
        ############################################################################################################################

        self.delimiter = delimiter

        import pandas as pd
        self.df = pd.read_csv   ( self.file
                                , delimiter = self.delimiter
                                , header = 0
                                )

        if self._isProgressFile:
            self._offset = -1
        else:
            self._offset = list(self.df.columns).index(self._sampleLogFuncColName) + 1 # index of the first variable
            self.ndim = len(self.df.columns) - self._offset

        self.count = len(self.df.iloc[:,1])
        self.ncol = len(self.df.iloc[1,:])

        if markovChainRequested:

            CumSumWeight = np.cumsum(self.df.iloc[:,self._offset-2].values, dtype=np.int32)
            if CumSumWeight[-1] != self.count: # it is indeed a compact chain
                #dfMarkov = pd.DataFrame( columns=list(self.df.columns), index=list(range(CumSumWeight[-1])) )
                dfMarkov = np.zeros( (CumSumWeight[-1] , self.ndim+self._offset) )
                istart = 0
                for i in range(self.count):
                    iend = CumSumWeight[i]
                    #dfMarkov.iloc[istart:iend,:] = self.df.iloc[i].values
                    dfMarkov[istart:iend,:] = self.df.iloc[i].values
                    istart = iend
                columns = self.df.columns
                self.df = pd.DataFrame(dfMarkov)
                self.count = len(self.df.iloc[:,1])
                self.df.columns = columns

        self._progress.note()
        if not self._isProgressFile:
            self._progress.note( msg = "ndim = " + str(self.ndim) + ", count = " + str(self.count), end = newline, pre = True )

        # set dynamic properties

        if parseContents:
            self._progress.note( msg = "parsing file contents... ", end = newline, pre = True )
            self.contents = Struct()
            for icol, colName in enumerate(self.df.columns):
                setattr ( self.contents, colName, self.df[colName] )

        ############################################################################################################################
        #### statistics
        ############################################################################################################################

        if not self._isProgressFile:

            self.stats = Struct()

            #### add chain cormat

            self._progress.note( msg = "computing the sample correlation matrix... ", end = newline, pre = True )
            self.stats.cormat = ccm.CorMat  ( dataFrame     = self.df
                                            , columns       = range(self._offset,self._offset+self.ndim)
                                            , methodName    = self._methodName
                                            , reportEnabled = self._reportEnabled
                                            , method        = "pearson"
                                            )
            self.stats.cormat()

            #### add chain covmat

            self._progress.note( msg = "computing the sample covariance matrix... ", end = newline, pre = True )
            self.stats.covmat = ccm.CovMat  ( dataFrame     = self.df
                                            , columns       = range(self._offset,self._offset+self.ndim)
                                            , methodName    = self._methodName
                                            , reportEnabled = self._reportEnabled
                                            )
            self.stats.covmat()

            #### add chain autocorrelation

            self._progress.note( msg = "computing the sample autocorrelations... ", end = newline, pre = True )
            self.stats.autocorr = AutoCorr  ( dataFrame     = self.df
                                            , columns       = range(self._offset-1,self._offset+self.ndim)
                                            , methodName    = self._methodName
                                            , reportEnabled = self._reportEnabled
                                            )
            self.stats.autocorr()

            #### add chain maxLogFunc

            self.stats.maxLogFunc = getMaxLogFunc(dataFrame = self.df)

        ############################################################################################################################
        #### graphics
        ############################################################################################################################

        self._plotTypeList =    [ "line"
                                , "scatter"
                                , "lineScatter"
                                ]

        if not self._isProgressFile: self._plotTypeList +=  [ "line3"
                                                            , "scatter3"
                                                            , "lineScatter3"
                                                            , "jointplot"
                                                            , "histplot"
                                                            , "kdeplot1"
                                                            , "kdeplot2"
                                                            , "contour3"
                                                            , "contourf"
                                                            , "contour"
                                                            , "grid"
                                                            ]

        self._progress.note( msg = "adding the graphics tools... ", end = newline, pre = True )
        self.plot = Struct()
        self._resetPlot(resetType="hard")
        self.plot.reset = self._resetPlot
        #self.plot.helpme = self.helpme

    ################################################################################################################################
    #### _resetPlot
    ################################################################################################################################

    def _resetPlot  ( self
                    , resetType = "soft"
                    , plotNames = "all"
                    ):
        """

        Reset the properties of the plot to the original default settings.
        Use this method when you change many attributes of the plot and
        you want to clean up and go back to the default settings.

            **Parameters**

                resetType (optional)

                    An optional string with possible value of ``"hard"``.
                    If provided, the plot object will be regenerated from scratch.
                    This includes reading the original data frame again and resetting
                    everything. If not provided, then only the plot settings will be
                    reset without reseting the dataFrame.

                plotNames (optional)

                    An optional string value or list of string values representing
                    the names of plots to reset. If no value is provided,
                    then all plots will be reset.

            **Returns**

                None

            **Example**

                .. code-block:: python

                    reset("hard")                   # regenerate all plots from scratch
                    reset("hard","line3")           # regenerate line3 plot from scratch
                    reset("hard",["line","line3"])  # regenerate line and line3 plots

        """

        requestedPlotTypeList = []
        if isinstance(plotNames, str):
            plotTypeLower = plotNames.lower()
            if plotTypeLower=="all":
                requestedPlotTypeList = self._plotTypeList
            elif plotNames in self._plotTypeList:
                requestedPlotTypeList = [plotNames]
            else:
                self._reportWrongPlotName(plotNames)
        elif isinstance(plotNames, list):
            for plotName in plotNames:
                if plotName not in self._plotTypeList: self._reportWrongPlotName(plotName)
        else:
            self._reportWrongPlotName("a none-string none-list object.")

        resetTypeIsHard = None
        if isinstance(resetType, str):
            resetTypeIsHard = resetType.lower()=="hard"
        else:
            resetTypeIsHard = None
            pm.abort( msg   = "The input argument resetType must be a string representing" + newline
                            + "the type of the reset to be performed on the plots." + newline
                            + "A list of possible plots includes: \"hard\", \"soft\"" + newline
                            + "Here is the help for the ``reset()`` method: " + newline
                            + newline
                            + self._resetPlot.__doc__
                    , marginTop = 1
                    , marginBot = 1
                    , methodName = self._methodName
                    )

        ############################################################################################################################
        #### reset plots
        ############################################################################################################################

        for requestedPlotType in requestedPlotTypeList:

            plotObject = None
            requestedPlotTypeLower = requestedPlotType.lower()

            is3d        = "3"           in requestedPlotTypeLower
            isLine      = "line"        in requestedPlotTypeLower
            isScatter   = "scatter"     in requestedPlotTypeLower

            isJointplot = "jointplot"   in requestedPlotTypeLower
            isHistplot  = "histplot"    in requestedPlotTypeLower
            isKdeplot1  = "kdeplot1"    in requestedPlotTypeLower
            isKdeplot2  = "kdeplot2"    in requestedPlotTypeLower
            isContourf  = "contourf"    in requestedPlotTypeLower
            isContour3  = "contour3"    in requestedPlotTypeLower
            isContour   = "contour"     in requestedPlotTypeLower and not (isContourf or isContour3)

            isGridPlot  = "grid"        in requestedPlotTypeLower

            isLineScatterPlot = isLine or isScatter
            isDensityPlot = isJointplot or isHistplot or isKdeplot1 or isKdeplot2 or isContourf or isContour3 or isContour

            if not resetTypeIsHard:
                plotComponent = getattr(self, "plot")
                plotObject = getattr(plotComponent, requestedPlotType)
                plotObject._reset()

            ########################################################################################################################
            #### reset line / scatter
            ########################################################################################################################

            if isLineScatterPlot:

                if resetTypeIsHard:
                    plotObject = LineScatterPlot( plotType = requestedPlotType
                                                , dataFrame = self.df
                                                , methodName = self._methodName
                                                , reportEnabled = self._reportEnabled
                                                , resetPlot = self._resetPlot
                                                )

                plotObject.ycolumns = self.df.columns[self._offset] # :]
                plotObject.ccolumns = self._sampleLogFuncColName
                plotObject.colorbar.kws.extend = "neither"
                plotObject.colorbar.kws.orientation = "vertical"
                plotObject.colorbar.kws.spacing = "uniform"

                if is3d:
                    plotObject.zcolumns = self._sampleLogFuncColName
                    if self.ndim>1:
                        plotObject.xcolumns = self.df.columns[self._offset]
                        plotObject.ycolumns = self.df.columns[self._offset+1]


                if isLine:
                    if isScatter:
                        plotObject.lineCollection.enabled = False
                        plotObject.plot.enabled = True
                        plotObject.plot.kws.alpha = 0.2
                        plotObject.plot.kws.color = "grey"
                        plotObject.plot.kws.linewidth = 0.75
                    else:
                        plotObject.lineCollection.enabled = True
                        plotObject.plot.enabled = False

            ########################################################################################################################
            #### reset density plots: kdeplot / histplot / jointplot / contour / contourf / contour3
            ########################################################################################################################

            if isDensityPlot:

                if resetTypeIsHard:
                    plotObject = DensityPlot( plotType = requestedPlotType
                                            , dataFrame = self.df
                                            , methodName = self._methodName
                                            , reportEnabled = self._reportEnabled
                                            , resetPlot = self._resetPlot
                                            )

                plotObject.xcolumns = self.df.columns[self._offset]
                if not (isHistplot or isKdeplot1):
                    if self.ndim==1:
                        plotObject.xcolumns = self.df.columns[self._offset-1]
                        plotObject.ycolumns = self.df.columns[self._offset]
                    else:
                        plotObject.ycolumns = self.df.columns[self._offset+1]

            ########################################################################################################################
            #### reset GridPlot
            ########################################################################################################################

            if isGridPlot:

                if resetTypeIsHard:
                    plotObject = GridPlot   ( plotType = requestedPlotType
                                            , dataFrame = self.df
                                            , methodName = self._methodName
                                            , reportEnabled = self._reportEnabled
                                            , resetPlot = self._resetPlot
                                            )

                endColindex = np.min( [self._offset+3, self._offset+self.ndim] )
                plotObject.columns = self.df.columns[self._offset-1:endColindex]
                plotObject.ccolumn = self._sampleLogFuncColName

            ########################################################################################################################
            #### reset target component
            ########################################################################################################################

            if (isLineScatterPlot or isDensityPlot) and not (plotObject._type.is3d or self._isProgressFile):

                xtarget = 0 # dummy
                if isDensityPlot: xtarget = self.df[plotObject.xcolumns].values.flatten()[self.stats.maxLogFunc.idrow]

                if plotObject._type.is1d: plotObject.target.value = [ xtarget, 0 ]

                if plotObject._type.is2d:
                    ytarget = self.df[plotObject.ycolumns].values.flatten()[self.stats.maxLogFunc.idrow]
                    plotObject.target.value = [ xtarget, ytarget ]

                if isDensityPlot and plotObject._type.is1d: plotObject.target.axhline.enabled = False
                if isLine or isScatter: plotObject.target.axvline.enabled = False
                plotObject.target.label = "maxLogFunc"

            ########################################################################################################################

            if plotObject is not None: setattr(self.plot, requestedPlotType, plotObject)

    ################################################################################################################################
    #### _reportWrongPlotName
    ################################################################################################################################

    def _reportWrongPlotName( self
                            , plotNames
                            ):

        pm.abort( msg   = "The input argument ``plotNames`` must be a string representing" + newline
                        + "the name of a plot belonging to the TabularFileContents class or," + newline
                        + "a list of such plot names. You have entered: " + plotNames + newline
                        + "Possible plots are: " + newline
                        + newline
                        + newline.join(self._plotTypeList) + newline
                        + newline
                        + "Here is the help for the ``reset()`` method: " + newline
                        + newline
                        + self._resetPlot.__doc__
                , marginTop = 1
                , marginBot = 1
                , methodName = self._methodName
                )

    ################################################################################################################################




####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
import typing as tp

from _ParaMonteSampler import ParaMonteSampler
from _TabularFileContents import TabularFileContents
import _paramonte as pm

newline = pm.newline

####################################################################################################################################
#### ParaDRAM class
####################################################################################################################################

class ParaDRAM(ParaMonteSampler):
    """

    This is the **ParaDRAM** class to generate instances of **serial** and **parallel**
    **Delayed-Rejection Adaptive Metropolis-Hastings Markov Chain Monte Carlo** 
    sampler class of the ParaMonte library. The ``ParaDRAM`` class is a 
    child of the ``ParaMonteSampler`` class.

    All ParaDRAM class attributes (input arguments to the ParaDRAM constructor)
    are optional and all attributes can be also set after a ParaDRAM instance
    is returned by the constructor.

    Once you set the optional attributes to your desired values,
    call the ParaDRAM sampler via the object's method ``runSampler()``.

    .. _example-serial-usage:

    **Example serial usage**

    Copy and paste the following code enclosed between the
    two comment lines in your python/ipython/jupyter session
    (ensure the indentations of the pasted lines comply with Python rules):

    .. code-block:: python
        :linenos:

        ##################################
        import paramonte as pm
        import numpy as np
        def getLogFunc(point):
            # return the log of the standard multivariate
            # Normal density function with ndim dimensions
            return -0.5 * np.sum( np.double( point )**2 )
        pmpd = pm.ParaDRAM()
        pmpd.runSampler ( ndim = 4 # assume 4-dimensional objective function
                        , getLogFunc = getLogFunc   # the objective function
                        )
        ##################################

    where,

        ndim

            represents the number of dimensions of the domain of
            the user's objective function ``getLogFunc(point)`` and,

        getLogFunc(point)

            represents the user's objective function to be sampled,
            which must take a single input argument ``point`` of type
            numpy-float64 array of length ``ndim`` and must return the
            natural logarithm of the objective function.

    .. _example-parallel-usage:

    **Example parallel usage**

    Copy and paste the following code enclosed between the
    two comment lines in your python/ipython/jupyter session
    (ensure the indentations of the pasted lines comply with Python rules):

    .. code-block:: python
        :linenos:

        ##################################
        with open("main.py", "w") as file:
            file.write  ('''
        import paramonte as pm
        import numpy as np
        def getLogFunc(point):
            # return the log of the standard multivariate
            # Normal density function with ndim dimensions
            return -0.5 * np.sum( np.double( point )**2 )
        pmpd = pm.ParaDRAM()
        pmpd.mpiEnabled = True
        pmpd.runSampler ( ndim = 4 # assume 4-dimensional objective function
                        , getLogFunc = getLogFunc   # the objective function
                        )
        ''')
        ##################################

    where,

        ndim

            represents the number of dimensions of the domain of
            the user's objective function ``getLogFunc(point)`` and,

        getLogFunc(point)

            represents the user's objective function that is to be sampled.
            This function must take a single input argument ``point`` of type
            numpy-float64 array of length ndim and must return the natural
            logarithm of the objective function.

        mpiEnabled

            is a logical (boolean) indicator that, if ``True``, will
            cause the ParaDRAM simulation to run in parallel
            on the requested number of processors.
            The default value is ``False``.

    The above will generate a Parallel-ParaDRAM-simulation Python script in the
    current working directory of Python. Note the only difference between the
    serial and parallel simulation scripts: the extra line ``pmpd.mpiEnabled = True``
    which tell the ParaMonte library to run the simulation in parallel.

    Assuming that you already have an MPI runtime library installed on your
    system (see below), you can now execute this Python script file ``main.py``
    in parallel in two ways:

    1.  from inside ipython or jupyter, type the following,

        .. code-block:: bash

            !mpiexec -n 3 python main.py

    2.  outside of Python environment,
        from within a Bash shell (on Linux or Mac) or,
        from within an Anaconda command prompt on Windows,
        type the following,

        .. code-block:: bash

            mpiexec -n 3 python main.py

    **Note:**

    On Windows platform, if you are using the Intel MPI library,
    we recommend that you also specify the extra flag -localonly,

    .. code-block:: bash

        mpiexec -localonly -n 3 python main.py

    This will cause the simulations to run in parallel only on a single node,
    but more importantly, it will also prevent the use of Hydra service and
    the requirement for its registration. If you are not on a Windows cluster,
    (e.g., you are using your personal device), then we highly recommend
    specifying this flag.


    In all cases in the above, the script ``main.py`` will run on 3 processors.
    Feel free to change the number of processors to any number desired. But do
    not request more than the available number of physical cores on your system.

    **Tips on parallel usage**

    For up-to-date detailed instructions on how to run simulations in parallel visit:

        https://www.cdslab.org/paramonte

    You can also use the following commands on the Python command-line,

    .. code-block:: python
        :linenos:

        ##################################
        import paramonte as pm
        pm.verify() # verify the existence of parallel simulation prerequisites
        ##################################

    to obtain specific information on how to run a parallel simulation,
    in particular, in relation to your current installation of ParaMonte.
    In general, for parallel simulations:

    0.  Ensure you need and will get a speedup by running the ParaDRAM sampler
        in parallel. Typically, if a single evaluation of the objective function
        takes much longer than a few milliseconds, your simulation may then
        benefit from the parallel run.

    1.  Ensure you have an MPI library installed, preferably, the Intel MPI
        runtime libraries. An MPI library should be automatically installed
        on your system with ParaMonte. If needed, you can download the Intel
        MPI library from their website and install it.

    2.  Ensure the ParaDRAM object property ``mpiEnabled`` is ``True``
        (the default is ``False``).

    3.  Before running the parallel simulation, in particular, on Windows systems,
        you may need to define the necessary MPI environmental variables on your system.
        To get information on how to define the variables, use the paramonte module's
        function, ``verify()``, as described in the above.

    4.  Call your main Python code from a Python-aware mpiexec-aware command-line via,

        .. code-block:: bash

            mpi_launcher -n num_process python name_of_yor_python_code.py

        where,

        1.  "mpi_launcher" is the name of the MPI launcher
            of the MPI runtime library that you have installed.
            For example, the Intel MPI library's launcher is named mpiexec,
            also recognized by Microsoft, MPICH, and OpenMPI.
            Note that on supercomputers, the MPI launcher is usually
            something other than ``mpiexec``, for example:
            ``ibrun``, ``mpirun``, ...

        2.  "num_process" represents the number of cores
            on which you want to run the program. Replace this
            with the an integer number, like, 3 (meaning 3 cores).

            Do not assign more processes than the available number of
            physical cores on your device/cluster. Assigning more cores
            than physically available on your system will only slow down
            your simulation.

    Once the above script is saved in the file ``main.py``, open a Python-aware and
    MPI-aware command prompt to run the simulation in parallel via the MPI launcher,

    .. code-block:: bash

        mpiexec -n 3 python main.py

    This will execute the Python script ``main.py`` on three processes (images).
    Keep in mind that on Windows systems you may need to define MPI environmental
    variables before a parallel simulation, as described in the above.

    **ParaDRAM Class Attributes**

    See also:

        https://www.cdslab.org/paramonte/notes/usage/paradram/specifications/

    All input specifications (attributes) of a ParaDRAM simulation are optional.
    However, it is recommended that you provide as much information as possible
    about the specific ParaDRAM simulation and the objective function to be sampled
    via ParaDRAM simulation specifications.

    The ParaDRAM simulation specifications have lengthy comprehensive descriptions
    that appear in full in the output report file of every ParaDRAM simulation.

    The best way to learn about individual ParaDRAM simulation attributes
    is to a run a minimal serial simulation with the following Python script,

    .. code-block:: python
        :linenos:

        ##################################
        from paramonte import ParaDRAM
        pmpd = ParaDRAM()
        pmpd.spec.outputFileName = "./test"
        def getLogFunc(point): return -sum(point**2)
        pmpd.runSampler( ndim = 1, getLogFunc = getLogFunc )
        ##################################

    Running this code will generate a set of simulation output files (in the current
    working directory of Python) that begin with the prefix ``test_process_1``. Among
    these, the file ``test_process_1_report.txt`` contains the full description of all
    input specifications of the ParaDRAM simulation as well as other information
    about the simulation results and statistics.

    **Parameters**

        None. The simulation specifications can be set once an object is instantiated.
        All simulation specification descriptions are collectively available at:

            https://www.cdslab.org/paramonte/notes/usage/paradram/specifications/

        Note that this is the new interface. The previous ParaDRAM class interface
        used to optionally take all simulation specifications as input. However,
        overtime, this approach has become more of liability than any potential
        benefit. All simulation specifications have to be now to be set solely
        after a ParaDRAM object is instantiated, instead of setting the
        specifications via the ParaDRAM class constructor.

    **Attributes**

        buildMode

            optional string argument with the default value "release".
            possible choices are:

                "debug"

                    to be used for identifying sources of bug
                    and causes of code crash.

                "release"

                    to be used in all other normal scenarios
                    for maximum runtime efficiency.

        mpiEnabled

            optional logical (boolean) indicator which is ``False`` by default.
            If it is set to ``True``, it will cause the ParaDRAM simulation
            to run in parallel on the requested number of processors.
            See the class documentation guidelines in the above for
            information on how to run a simulation in parallel.

        reportEnabled

            optional logical (boolean) indicator which is ``True`` by default.
            If it is set to ``True``, it will cause extensive guidelines to be
            printed on the standard output as the simulation or post-processing
            continues with hints on the next possible steps that could be taken
            in the process. If you do not need such help and information set
            this variable to ``False`` to silence all output messages.

        inputFile

            optional string input representing the path to
            an external input namelist of simulation specifications.
            USE THIS OPTIONAL ARGUMENT WITH CAUTION AND
            ONLY IF YOU KNOW WHAT YOU ARE DOING.

            **WARNING**

            Specifying an input file will cause the ParaDRAM sampler
            to ignore all other simulation specifications set by the 
            user via sampler instance's `spec`-component attributes.

        spec

            A frozen class containing all simulation specifications.
            All simulation attributes are by default set to appropriate
            values at runtime. To override the default simulation
            specifications, set the `spec` attributes to some
            desired values of your choice. For possible values, see:

                https://www.cdslab.org/paramonte/notes/usage/paradram/specifications/

            If you need help on any of the simulation specifications, try
            the supplied ``helpme()`` function in this component, like,

            .. code-block:: python
                :linenos:

                ##################################
                import paramonte as pm
                pmpd = pm.ParaDRAM()          # instantiate a ParaDRAM sampler class
                pmpd.spec.helpme()            # get help on all simulation specification
                pmpd.spec.helpme("chainSize") # get help on "chainSize" specifically
                ##################################

    **Methods**

        See below for information on the methods.

    **Returns**

        Object of class ParaDRAM sampler.

    ---------------------------------------------------------------------------
    """

    def __init__(self):
        """

        The constructor for ParaDRAM class.
        All input parameters are optional and all class
        attributes can be changed after the object construction.

            **Parameters**

                None

        """

        super().__init__(methodName = "ParaDRAM")

        ## ParaMonte specifications
        #
        #self.spec = pm.utils.FrozenClass()
        #
        ## ParaMonte variables
        #self.spec.sampleSize                           = sampleSize
        #self.spec.randomSeed                           = randomSeed
        #self.spec.description                          = description
        #self.spec.outputFileName                       = outputFileName
        #self.spec.outputDelimiter                      = outputDelimiter
        #self.spec.chainFileFormat                      = chainFileFormat
        #self.spec.variableNameList                     = variableNameList
        #self.spec.restartFileFormat                    = restartFileFormat
        #self.spec.outputColumnWidth                    = outputColumnWidth
        #self.spec.outputRealPrecision                  = outputRealPrecision
        #self.spec.silentModeRequested                  = silentModeRequested
        #self.spec.domainLowerLimitVec                  = domainLowerLimitVec
        #self.spec.domainUpperLimitVec                  = domainUpperLimitVec
        #self.spec.parallelizationModel                 = parallelizationModel
        #self.spec.progressReportPeriod                 = progressReportPeriod
        #self.spec.targetAcceptanceRate                 = targetAcceptanceRate
        #self.spec.mpiFinalizeRequested                 = mpiFinalizeRequested
        #self.spec.maxNumDomainCheckToWarn              = maxNumDomainCheckToWarn
        #self.spec.maxNumDomainCheckToStop              = maxNumDomainCheckToStop
        ## ParaMCMC variables
        #self.spec.chainSize                            = chainSize
        #self.spec.scaleFactor                          = scaleFactor
        #self.spec.startPointVec                        = startPointVec
        #self.spec.proposalModel                        = proposalModel
        #self.spec.proposalStartCovMat                  = proposalStartCovMat
        #self.spec.proposalStartCorMat                  = proposalStartCorMat
        #self.spec.proposalStartStdVec                  = proposalStartStdVec
        #self.spec.sampleRefinementCount                = sampleRefinementCount
        #self.spec.sampleRefinementMethod               = sampleRefinementMethod
        #self.spec.randomStartPointRequested            = randomStartPointRequested
        #self.spec.randomStartPointDomainLowerLimitVec  = randomStartPointDomainLowerLimitVec
        #self.spec.randomStartPointDomainUpperLimitVec  = randomStartPointDomainUpperLimitVec
        ## ParaDRAM variables
        #self.spec.adaptiveUpdateCount                  = adaptiveUpdateCount
        #self.spec.adaptiveUpdatePeriod                 = adaptiveUpdatePeriod
        #self.spec.greedyAdaptationCount                = greedyAdaptationCount
        #self.spec.delayedRejectionCount                = delayedRejectionCount
        #self.spec.burninAdaptationMeasure              = burninAdaptationMeasure
        #self.spec.delayedRejectionScaleFactorVec       = delayedRejectionScaleFactorVec
        #
        #self.spec.helpme = SpecDRAM.helpme
        #self.spec._freeze()

    ################################################################################################################################
    #### runSampler
    ################################################################################################################################

    def runSampler  ( self
                    , ndim          : int
                    , getLogFunc    : tp.Callable[[tp.List[float]], float]
                    , inputFile     : tp.Optional[str] = None
                    ) -> None:
        """

        Run ParaDRAM sampler and return nothing.

            **Parameters**

                ndim

                    An integer representing the number of dimensions of the
                    domain of the user's objective function ``getLogFunc(point)``.
                    It must be a positive integer.

                getLogFunc(point)

                    represents the user's objective function to be sampled,
                    which must take a single input argument ``point`` of type
                    numpy-float64 array of length ``ndim`` and must return the
                    natural logarithm of the objective function.

                inputFile (optional)

                    A string input representing the path to an external
                    input namelist of simulation specifications.

                        **WARNING**

                        Use this optional argument with caution and only
                        if you know what you are doing. Specifying this option
                        will cause the sampler to ignore all other simulation
                        specifications set by the user via the ``spec``
                        component of the sampler instance.

            **Returns**

                None

        """

        if not isinstance(ndim,int) or ndim<1:
            pm.abort( msg   = "The input argument ndim must be a positive integer," + newline
                            + "representing the number of dimensions of the domain of" + newline
                            + "the user's objective function getLogFunc()." + newline
                            + "You have entered ndim = " + str(ndim)
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        if not callable(getLogFunc):
            pm.abort( msg   = "The input argument getLogFunc must be a callable function." + newline
                            + "It represents the user's objective function to be sampled," + newline
                            + "which must take a single input argument of type numpy" + newline
                            + "float64 array of length ndim and must return the" + newline
                            + "natural logarithm of the objective function."
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        if inputFile is not None and not isinstance(inputFile,str):
            pm.abort( msg   = "The input argument ``inputFile`` must be of type str." + newline
                            + "It is an optional string input representing the path to" + newline
                            + "an external input namelist of simulation specifications." + newline
                            + "USE THIS OPTIONAL ARGUMENT WITH CAUTION AND" + newline
                            + "ONLY IF YOU KNOW WHAT YOU ARE DOING." + newline
                            + "Specifying this option will cause the sampler to ignore" + newline
                            + "all other simulation specifications set by the user via" + newline
                            + "the ``spec`` component of the sampler instance." + newline
                            + "You have entered inputFile = " + str(inputFile)
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        def getLogFunc2arg(ndim,point):
            PointVec = np.array(point[0:ndim])
            return getLogFunc(PointVec)

        self._runSampler( ndim
                        , getLogFunc2arg
                        , inputFile
                        )

    ################################################################################################################################
    #### readMarkovChain
    ################################################################################################################################

    def readMarkovChain ( self
                        , file          : tp.Optional[str] = None
                        , delimiter     : tp.Optional[str] = None
                        , parseContents : tp.Optional[bool] = True
                        , renabled      : tp.Optional[bool] = False
                        ) -> tp.List[TabularFileContents] :
        """

        Return a list of the unweighted verbose (Markov-chain) contents
        of a set of ParaDRAM output chain files, whose names begin the
        user-provided input variable ``file``. This method is to be only
        used for the postprocessing of the output chain file(s) of an
        already finished ParaDRAM simulation. It is not meant to be called
        by all processes in parallel mode, although it is possible.

            **Parameters**

                file (optional)

                    A string representing the path to the chain file with
                    the default value of ``None``.
                    The path only needs to uniquely identify the simulation
                    to which the chain file belongs. For example, specifying
                    ``"./mydir/mysim"`` as input will lead to a search for a file
                    that begins with ``"mysim"`` and ends with ``"_chain.txt"``
                    inside the directory ``"./mydir/"``. If there are multiple
                    files with such name, then all of them will be read
                    and returned as a list.
                    If this input argument is not provided by the user, the
                    value of the object attribute ``outputFileName`` will be
                    used instead. At least one of the two mentioned routes
                    must provide the path to the chain file otherwise,
                    this method will break by calling ``sys.exit()``.

                delimiter (optional)

                    An input string representing the delimiter used in the
                    output chain file. If it is not provided as input argument, the
                    value of the corresponding object attribute ``outputDelimiter``
                    will be used instead. If none of the two are available,
                    the default comma delimiter ``","`` will be assumed and used.

                parseContents (optional)

                    If set to ``True``, the contents of the file will be parsed
                    and stored in a component of the object named ``contents``.
                    The default value is ``True``.

                renabled (optional)

                    If set to False, the contents of the file(s) will be stored
                    as a list in a (new) component of the ParaDRAM object named
                    ``markovChainList`` and ``None`` will be the return value
                    of the method. If set to True, the reverse will done.
                    The default value is ``False``.

            **Returns**

                A list of objects, each of which has the following properties:

                    file

                        The full absolute path to the chain file.

                    delimiter

                        The delimiter used in the chain file.

                    ndim

                        The number of dimensions of the domain of the objective
                        function from which the chain has been drawn.

                    count

                        The number of unique (weighted) points in the chain file.
                        This is essentially the number of rows in the chain file
                        minus one (representing the header line).

                    plot

                        A structure containing the graphics tools for the 
                        visualization of the contents of the file.

                    df

                        The unweighted (Markovian) contents of the chain file in the
                        form of a pandas-library DataFrame (hence called ``df``).

                    contents

                        corresponding to each column in the progress file, a property
                        with the same name as the column header is also created
                        for the object which contains the data stored in that column
                        of the progress file. These properties are all stored in the 
                        attribute ``contents``.

                If ``renabled = True``, the list of objects will be returned as the
                return value of the method. Otherwise, the list will be stored in a
                component of the ParaDRAM object named ``markovChainList``.

        """

        return self._readTabular( file = file
                                , fileType = "markovChain"
                                , delimiter = delimiter
                                , parseContents = parseContents
                                , renabled = renabled
                                )

    ################################################################################################################################

